@@ -1,9 +1,9 @@
 #! /bin/sh
 # check script for Pdlzip - LZMA lossless data compressor
-# Copyright (C) 2010-2019 Antonio Diaz Diaz.
+# Copyright (C) 2010-2021 Antonio Diaz Diaz.
 #
 # This script is free software: you have unlimited permission
-# to copy, distribute and modify it.
+# to copy, distribute, and modify it.
 
 LC_ALL=C
 export LC_ALL
@@ -30,6 +30,8 @@ cd "${objdir}"/tmp || framework_failure
 
 cat "${testdir}"/test.txt > in || framework_failure
 in_lz="${testdir}"/test.txt.lz
+in_em="${testdir}"/test_em.txt.lz
+fox_lz="${testdir}"/fox.lz
 fail=0
 test_failed() { fail=1 ; printf " $1" ; [ -z "$2" ] || printf "($2)" ; }
 
@@ -56,8 +58,20 @@ done
 [ $? = 2 ] || test_failed $LINENO
 "${LZIP}" -dq -o in < "${in_lz}"
 [ $? = 1 ] || test_failed $LINENO
+"${LZIP}" -dq -o in "${in_lz}"
+[ $? = 1 ] || test_failed $LINENO
+"${LZIP}" -dq -o out nx_file.lz
+[ $? = 1 ] || test_failed $LINENO
+[ ! -e out ] || test_failed $LINENO
+"${LZIP}" -q -o out.lz nx_file
+[ $? = 1 ] || test_failed $LINENO
+[ ! -e out.lz ] || test_failed $LINENO
 # these are for code coverage
-"${LZIP}" -t -- nx_file 2> /dev/null
+"${LZIP}" -cdt "${in_lz}" > out 2> /dev/null
+[ $? = 1 ] || test_failed $LINENO
+"${LZIP}" -t -- nx_file.lz 2> /dev/null
+[ $? = 1 ] || test_failed $LINENO
+"${LZIP}" -t "" < /dev/null 2> /dev/null
 [ $? = 1 ] || test_failed $LINENO
 "${LZIP}" --help > /dev/null || test_failed $LINENO
 "${LZIP}" -n1 -V > /dev/null || test_failed $LINENO
@@ -81,15 +95,22 @@ printf "LZIP\001+.............................." | "${LZIP}" -t 2> /dev/null
 
 printf "\ntesting decompression..."
 
-"${LZIP}" -t "${in_lz}" || test_failed $LINENO
-"${LZIP}" -cd "${in_lz}" > copy || test_failed $LINENO
-cmp in copy || test_failed $LINENO
+for i in "${in_lz}" "${in_em}" "${testdir}"/test.txt.lzma ; do
+	"${LZIP}" -t "$i" || test_failed $LINENO "$i"
+	"${LZIP}" -d "$i" -o copy || test_failed $LINENO "$i"
+	cmp in copy || test_failed $LINENO "$i"
+	"${LZIP}" -cd "$i" > copy || test_failed $LINENO "$i"
+	cmp in copy || test_failed $LINENO "$i"
+	"${LZIP}" -d "$i" -o - > copy || test_failed $LINENO "$i"
+	cmp in copy || test_failed $LINENO "$i"
+	"${LZIP}" -d < "$i" > copy || test_failed $LINENO "$i"
+	cmp in copy || test_failed $LINENO "$i"
+	rm -f copy || framework_failure
+done
 
-"${LZIP}" -t "${testdir}"/test.txt.lzma || test_failed $LINENO
-"${LZIP}" -cd "${testdir}"/test.txt.lzma > copy || test_failed $LINENO
-cmp in copy || test_failed $LINENO
+lines=$("${LZIP}" -tvv "${in_em}" 2>&1 | wc -l) || test_failed $LINENO
+[ "${lines}" -eq 8 ] || test_failed $LINENO "${lines}"
 
-rm -f copy || framework_failure
 cat "${in_lz}" > copy.lz || framework_failure
 "${LZIP}" -dk copy.lz || test_failed $LINENO
 cmp in copy || test_failed $LINENO
@@ -100,19 +121,21 @@ printf "to be overwritten" > copy || framework_failure
 [ ! -e copy.lz ] || test_failed $LINENO
 cmp in copy || test_failed $LINENO
 
-rm -f copy || framework_failure
-cat "${in_lz}" > copy.lz || framework_failure
-"${LZIP}" -d -S100k copy.lz || test_failed $LINENO	# ignore -S
-[ ! -e copy.lz ] || test_failed $LINENO
-cmp in copy || test_failed $LINENO
-
 printf "to be overwritten" > copy || framework_failure
+"${LZIP}" -d -o copy < "${in_lz}" 2> /dev/null
+[ $? = 1 ] || test_failed $LINENO
 "${LZIP}" -df -o copy < "${in_lz}" || test_failed $LINENO
 cmp in copy || test_failed $LINENO
+rm -f out copy || framework_failure
+"${LZIP}" -d -o ./- "${in_lz}" || test_failed $LINENO
+cmp in ./- || test_failed $LINENO
+rm -f ./- || framework_failure
+"${LZIP}" -d -o ./- < "${in_lz}" || test_failed $LINENO
+cmp in ./- || test_failed $LINENO
+rm -f ./- || framework_failure
 
-rm -f copy || framework_failure
-"${LZIP}" -s16 < in > anyothername || test_failed $LINENO
-"${LZIP}" -dv --output copy - anyothername - < "${in_lz}" 2> /dev/null ||
+cat "${in_lz}" > anyothername || framework_failure
+"${LZIP}" -dv - anyothername - < "${in_lz}" > copy 2> /dev/null ||
 	test_failed $LINENO
 cmp in copy || test_failed $LINENO
 cmp in anyothername.out || test_failed $LINENO
@@ -147,19 +170,18 @@ done
 cmp in copy || test_failed $LINENO
 
 cat in in > in2 || framework_failure
-cat "${in_lz}" "${in_lz}" > in2.lz || framework_failure
-"${LZIP}" -t in2.lz || test_failed $LINENO
-"${LZIP}" -cd in2.lz > copy2 || test_failed $LINENO
+"${LZIP}" -t "${in_lz}" "${in_lz}" || test_failed $LINENO
+"${LZIP}" -cd "${in_lz}" "${in_lz}" -o out > copy2 || test_failed $LINENO
+[ ! -e out ] || test_failed $LINENO			# override -o
 cmp in2 copy2 || test_failed $LINENO
-
-"${LZIP}" -s16 --output=copy2.lz < in2 || test_failed $LINENO
-"${LZIP}" -t copy2.lz || test_failed $LINENO
-"${LZIP}" -cd copy2.lz > copy2 || test_failed $LINENO
+rm -f copy2 || framework_failure
+"${LZIP}" -d "${in_lz}" "${in_lz}" -o copy2 || test_failed $LINENO
 cmp in2 copy2 || test_failed $LINENO
+rm -f copy2 || framework_failure
 
+cat "${in_lz}" "${in_lz}" > copy2.lz || framework_failure
 printf "\ngarbage" >> copy2.lz || framework_failure
 "${LZIP}" -tvvvv copy2.lz 2> /dev/null || test_failed $LINENO
-rm -f copy2 || framework_failure
 "${LZIP}" -atq copy2.lz
 [ $? = 2 ] || test_failed $LINENO
 "${LZIP}" -atq < copy2.lz
@@ -179,37 +201,46 @@ printf "\ntesting   compression..."
 
 "${LZIP}" -cf "${in_lz}" > out 2> /dev/null	# /dev/null is a tty on OS/2
 [ $? = 1 ] || test_failed $LINENO
-"${LZIP}" -cFvvm36 -s16 "${in_lz}" > out 2> /dev/null || test_failed $LINENO
+"${LZIP}" -Fvvm36 -o - -s16 "${in_lz}" > out 2> /dev/null || test_failed $LINENO
 "${LZIP}" -cd out | "${LZIP}" -d > copy || test_failed $LINENO
 cmp in copy || test_failed $LINENO
 
+"${LZIP}" -0 -o ./- in || test_failed $LINENO
+"${LZIP}" -cd ./- | cmp in - || test_failed $LINENO
+rm -f ./- || framework_failure
+"${LZIP}" -0 -o ./- < in || test_failed $LINENO			# add .lz
+[ ! -e ./- ] || test_failed $LINENO
+"${LZIP}" -cd -- -.lz | cmp in - || test_failed $LINENO
+rm -f ./-.lz || framework_failure
+
 for i in s4Ki 0 1 2 3 4 5 6 7 8 9 ; do
 	"${LZIP}" -k -$i -s16 in || test_failed $LINENO $i
 	mv -f in.lz copy.lz || test_failed $LINENO $i
 	printf "garbage" >> copy.lz || framework_failure
 	"${LZIP}" -df copy.lz || test_failed $LINENO $i
 	cmp in copy || test_failed $LINENO $i
-done
 
-for i in s4Ki 0 1 2 3 4 5 6 7 8 9 ; do
-	"${LZIP}" -c -$i -s16 in > out || test_failed $LINENO $i
+	"${LZIP}" -$i -s16 in -c > out || test_failed $LINENO $i
+	"${LZIP}" -$i -s16 in -o o_out || test_failed $LINENO $i # don't add .lz
+	[ ! -e o_out.lz ] || test_failed $LINENO
+	cmp out o_out || test_failed $LINENO $i
+	rm -f o_out || framework_failure
 	printf "g" >> out || framework_failure
 	"${LZIP}" -cd out > copy || test_failed $LINENO $i
 	cmp in copy || test_failed $LINENO $i
-done
 
-for i in s4Ki 0 1 2 3 4 5 6 7 8 9 ; do
 	"${LZIP}" -$i -s16 < in > out || test_failed $LINENO $i
 	"${LZIP}" -d < out > copy || test_failed $LINENO $i
 	cmp in copy || test_failed $LINENO $i
-done
 
-for i in s4Ki 0 1 2 3 4 5 6 7 8 9 ; do
-	"${LZIP}" -f -$i -s16 -o out < in || test_failed $LINENO $i
+	rm -f out || framework_failure
+	printf "to be overwritten" > out.lz || framework_failure
+	"${LZIP}" -f -$i -s16 -o out < in || test_failed $LINENO $i  # add .lz
+	[ ! -e out ] || test_failed $LINENO
 	"${LZIP}" -df -o copy < out.lz || test_failed $LINENO $i
 	cmp in copy || test_failed $LINENO $i
 done
-rm -f out.lz || framework_failure
+rm -f out out.lz || framework_failure
 
 printf "\ntesting bad input..."
 
@@ -258,6 +289,21 @@ else
 fi
 rm -f int.lz || framework_failure
 
+for i in fox_v2.lz fox_s11.lz fox_de20.lz \
+         fox_bcrc.lz fox_crc0.lz fox_das46.lz fox_mes81.lz ; do
+	"${LZIP}" -tq "${testdir}"/$i
+	[ $? = 2 ] || test_failed $LINENO $i
+done
+
+"${LZIP}" -cd "${fox_lz}" > fox || test_failed $LINENO
+for i in fox_bcrc.lz fox_crc0.lz fox_das46.lz fox_mes81.lz ; do
+	"${LZIP}" -cdq "${testdir}"/$i > out
+	[ $? = 2 ] || test_failed $LINENO $i
+	cmp fox out || test_failed $LINENO $i
+done
+rm -f fox out || framework_failure
+
+cat "${in_lz}" "${in_lz}" > in2.lz || framework_failure
 cat "${in_lz}" "${in_lz}" "${in_lz}" > in3.lz || framework_failure
 if dd if=in3.lz of=trunc.lz bs=14752 count=1 2> /dev/null &&
    [ -e trunc.lz ] && cmp in2.lz trunc.lz > /dev/null 2>&1 ; then
@@ -280,13 +326,21 @@ rm -f in2.lz in3.lz trunc.lz out || framework_failure
 cat "${in_lz}" > ingin.lz || framework_failure
 printf "g" >> ingin.lz || framework_failure
 cat "${in_lz}" >> ingin.lz || framework_failure
+"${LZIP}" -atq ingin.lz
+[ $? = 2 ] || test_failed $LINENO
+"${LZIP}" -atq < ingin.lz
+[ $? = 2 ] || test_failed $LINENO
+"${LZIP}" -acdq ingin.lz > out
+[ $? = 2 ] || test_failed $LINENO
+"${LZIP}" -adq < ingin.lz > out
+[ $? = 2 ] || test_failed $LINENO
 "${LZIP}" -t ingin.lz || test_failed $LINENO
+"${LZIP}" -t < ingin.lz || test_failed $LINENO
 "${LZIP}" -cd ingin.lz > copy || test_failed $LINENO
 cmp in copy || test_failed $LINENO
-"${LZIP}" -t < ingin.lz || test_failed $LINENO
 "${LZIP}" -d < ingin.lz > copy || test_failed $LINENO
 cmp in copy || test_failed $LINENO
-rm -f copy ingin.lz || framework_failure
+rm -f copy ingin.lz out || framework_failure
 
 echo
 if [ ${fail} = 0 ] ; then
